# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

import datetime as dt
import zoneinfo

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango
from nbxmpp.modules.vcard4 import AdrProperty
from nbxmpp.modules.vcard4 import BDayProperty
from nbxmpp.modules.vcard4 import CaladruriProperty
from nbxmpp.modules.vcard4 import CaluriProperty
from nbxmpp.modules.vcard4 import CategoriesProperty
from nbxmpp.modules.vcard4 import ClientpidmapProperty
from nbxmpp.modules.vcard4 import EmailProperty
from nbxmpp.modules.vcard4 import FBurlProperty
from nbxmpp.modules.vcard4 import FnProperty
from nbxmpp.modules.vcard4 import GenderProperty
from nbxmpp.modules.vcard4 import GeoProperty
from nbxmpp.modules.vcard4 import ImppProperty
from nbxmpp.modules.vcard4 import KeyProperty
from nbxmpp.modules.vcard4 import LangProperty
from nbxmpp.modules.vcard4 import LogoProperty
from nbxmpp.modules.vcard4 import MemberProperty
from nbxmpp.modules.vcard4 import NicknameProperty
from nbxmpp.modules.vcard4 import NoteProperty
from nbxmpp.modules.vcard4 import NProperty
from nbxmpp.modules.vcard4 import OrgProperty
from nbxmpp.modules.vcard4 import Parameters
from nbxmpp.modules.vcard4 import PhotoProperty
from nbxmpp.modules.vcard4 import ProdidProperty
from nbxmpp.modules.vcard4 import PronounsProperty
from nbxmpp.modules.vcard4 import RelatedProperty
from nbxmpp.modules.vcard4 import RevProperty
from nbxmpp.modules.vcard4 import RoleProperty
from nbxmpp.modules.vcard4 import SoundProperty
from nbxmpp.modules.vcard4 import SourceProperty
from nbxmpp.modules.vcard4 import TelProperty
from nbxmpp.modules.vcard4 import TitleProperty
from nbxmpp.modules.vcard4 import TzProperty
from nbxmpp.modules.vcard4 import UidProperty
from nbxmpp.modules.vcard4 import UrlProperty
from nbxmpp.modules.vcard4 import VCard

from gajim.common import app
from gajim.common.i18n import _
from gajim.common.i18n import p_
from gajim.common.iana import get_zone_data
from gajim.common.iana.time_zones import ZONES
from gajim.common.util.text import escape_iri_path_segment
from gajim.common.util.uri import InvalidUri
from gajim.common.util.uri import parse_uri
from gajim.common.util.uri import Uri

from gajim.gtk.dropdown import GajimDropDown
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import convert_py_to_glib_datetime
from gajim.gtk.util.misc import open_uri

PropertyT = (
    AdrProperty
    | BDayProperty
    | CaladruriProperty
    | CaluriProperty
    | CategoriesProperty
    | ClientpidmapProperty
    | EmailProperty
    | FBurlProperty
    | FnProperty
    | GenderProperty
    | GeoProperty
    | ImppProperty
    | KeyProperty
    | LangProperty
    | LogoProperty
    | MemberProperty
    | NicknameProperty
    | NoteProperty
    | NProperty
    | OrgProperty
    | PhotoProperty
    | ProdidProperty
    | PronounsProperty
    | RelatedProperty
    | RevProperty
    | RoleProperty
    | SoundProperty
    | SourceProperty
    | TelProperty
    | TitleProperty
    | TzProperty
    | UidProperty
    | UrlProperty
)

SupportedPropertiesT = (
    AdrProperty
    | BDayProperty
    | EmailProperty
    | FnProperty
    | GenderProperty
    | ImppProperty
    | KeyProperty
    | NicknameProperty
    | NoteProperty
    | NProperty
    | OrgProperty
    | PronounsProperty
    | RoleProperty
    | TelProperty
    | TitleProperty
    | TzProperty
    | UrlProperty
)

TextEntryPropertiesT = (
    EmailProperty
    | FnProperty
    | ImppProperty
    | OrgProperty
    | PronounsProperty
    | RoleProperty
    | TelProperty
    | TitleProperty
    | UrlProperty
)

LABEL_DICT = {
    "fn": _("Full Name"),
    "n": _("Name"),
    "bday": _("Birthday"),
    "gender": _("Gender"),
    "pronouns": _("Pronouns"),
    "adr": p_("Profile", "Address"),
    "tel": _("Phone No."),
    "email": _("Email"),
    "impp": p_("Profile", "IM Address"),
    "title": p_("Profile", "Title"),
    "role": p_("Profile", "Role"),
    "org": p_("Profile", "Organisation"),
    "note": p_("Profile", "Note"),
    "url": _("URL"),
    "key": p_("Profile", "Public Encryption Key"),
    "tz": p_("Profile", "Timezone"),
}


FIELD_TOOLTIPS = {"key": _("Your public key or authentication certificate")}


ADR_FIELDS = ["street", "ext", "pobox", "code", "locality", "region", "country"]


ADR_PLACEHOLDER_TEXT = {
    "pobox": _("Post Office Box"),
    "street": _("Street"),
    "ext": _("Extended Address"),
    "locality": _("City"),
    "region": _("State"),
    "code": _("Postal Code"),
    "country": _("Country"),
}


DEFAULT_KWARGS: dict[str, dict[str, str | list[Any]]] = {
    "fn": {"value": ""},
    "bday": {"value": "", "value_type": "date"},
    "gender": {"sex": "", "identity": ""},
    "pronouns": {"value": ""},
    "adr": {},
    "email": {"value": ""},
    "impp": {"value": ""},
    "tel": {"value": "", "value_type": "text"},
    "org": {"values": []},
    "title": {"value": ""},
    "role": {"value": ""},
    "url": {"value": ""},
    "key": {"value": "", "value_type": "text"},
    "note": {"value": ""},
    "tz": {"value": "Europe/Amsterdam", "value_type": "text"},
}


PROPERTIES_WITH_TYPE = [
    "adr",
    "email",
    "impp",
    "tel",
    "key",
]


ORDER = [
    "fn",
    "gender",
    "pronouns",
    "bday",
    "adr",
    "email",
    "impp",
    "tel",
    "org",
    "title",
    "role",
    "tz",
    "url",
    "key",
    "note",
]


SEX_VALUES = {
    "M": _("Male"),
    "F": _("Female"),
    "O": p_("Gender", "Other"),
    "N": p_("Gender", "None"),
    "U": _("Unknown"),
}


TYPE_VALUES = {"-": None, "home": _("Home"), "work": _("Work")}


class VCardGrid(Gtk.Grid):
    def __init__(self, account: str) -> None:
        Gtk.Grid.__init__(
            self,
            row_spacing=12,
            column_spacing=12,
            halign=Gtk.Align.CENTER,
            width_request=300,
        )

        self._callbacks = {
            "fn": TextEntryPropertyGui,
            "bday": DatePropertyGui,
            "gender": GenderPropertyGui,
            "pronouns": TextEntryPropertyGui,
            "adr": AdrPropertyGui,
            "tel": TextEntryPropertyGui,
            "email": TextEntryPropertyGui,
            "impp": TextEntryPropertyGui,
            "title": TextEntryPropertyGui,
            "role": TextEntryPropertyGui,
            "org": TextEntryPropertyGui,
            "url": TextEntryPropertyGui,
            "key": KeyPropertyGui,
            "note": MultiLinePropertyGui,
            "tz": TzPropertyGui,
        }

        self._account = account
        self._row_count: int = 0
        self._vcard: VCard | None = None
        self._props: list[VCardPropertyGui] = []

    def run_destroy(self) -> None:
        self.clear()
        app.check_finalize(self)

    def set_editable(self, enabled: bool) -> None:
        for prop in self._props:
            prop.set_editable(enabled)

    def set_vcard(self, vcard: VCard) -> None:
        self.clear()
        self._vcard = vcard

        for entry in ORDER:
            for prop in cast(list[PropertyT], vcard.get_properties()):
                if entry != prop.name:
                    continue

                self.add_property(prop)

    def get_vcard(self) -> VCard | None:
        return self._vcard

    def validate(self) -> None:
        for prop in list(self._props):
            base_prop = prop.get_base_property()
            if base_prop.is_empty:
                self.remove_property(prop)

    def add_new_property(self, name: str) -> None:
        assert self._vcard is not None
        kwargs = DEFAULT_KWARGS[name]
        prop = cast(PropertyT, self._vcard.add_property(name, **kwargs))
        self.add_property(prop, editable=True)

    def add_property(self, prop: PropertyT, editable: bool = False) -> None:
        prop_class = self._callbacks.get(prop.name)
        if prop_class is None:
            return

        prop_obj = prop_class(prop, self._account)  # type: ignore
        prop_obj.set_editable(editable)
        prop_obj.add_to_grid(self, self._row_count)

        self._props.append(prop_obj)
        self._row_count += 1

    def remove_property(self, prop: VCardPropertyGui) -> None:
        self.remove_row(prop.row_number)
        self._props.remove(prop)
        assert self._vcard is not None
        self._vcard.remove_property(prop.get_base_property())

    def get_row_count(self) -> int:
        return self._row_count

    def clear(self) -> None:
        self._vcard = None
        self._row_count = 0
        for prop in list(self._props):
            self.remove_row(prop.row_number)
            prop.destroy()

        self._props = []

    def sort(self) -> None:
        assert self._vcard is not None
        self.set_vcard(self._vcard)


class DescriptionLabel(Gtk.Label):
    def __init__(self, value: str) -> None:
        Gtk.Label.__init__(self, label=LABEL_DICT[value])
        if value == "adr":
            self.set_valign(Gtk.Align.START)
        else:
            self.set_valign(Gtk.Align.CENTER)
        self.add_css_class("dimmed")
        self.add_css_class("me-18")
        self.set_visible(True)
        self.set_xalign(1)
        self.set_tooltip_text(FIELD_TOOLTIPS.get(value, ""))


class ValueLabel(Gtk.Label, SignalManager):
    def __init__(self, prop: TextEntryPropertiesT | BDayProperty, account: str) -> None:
        Gtk.Label.__init__(
            self,
            selectable=True,
            xalign=0,
            max_width_chars=50,
            wrap=True,
            wrap_mode=Pango.WrapMode.WORD_CHAR,
            valign=Gtk.Align.CENTER,
            halign=Gtk.Align.START,
        )
        SignalManager.__init__(self)

        self._prop = prop
        self._uri: Uri | None = None
        self._account = account

        self._connect(self, "activate-link", self._on_activate_link)
        if isinstance(prop, OrgProperty):
            self.set_value(prop.values[0] if prop.values else "")
        else:
            self.set_value(prop.value)

    def run_destroy(self) -> None:
        self._disconnect_all()
        app.check_finalize(self)

    def set_value(self, value: str) -> None:
        # EMAIL https://rfc-editor.org/rfc/rfc6350#section-6.4.2
        if self._prop.name == "email":
            uri = "mailto:" + escape_iri_path_segment(value)
            self.set_value_with_uri(value, uri)

        # All the properties that can be (single) URIs:
        # SOURCE    https://rfc-editor.org/rfc/rfc6350#section-6.1.3
        # PHOTO     https://rfc-editor.org/rfc/rfc6350#section-6.2.4
        # TEL       https://rfc-editor.org/rfc/rfc6350#section-6.4.1
        # IMPP      https://rfc-editor.org/rfc/rfc6350#section-6.4.3
        # TZ        https://rfc-editor.org/rfc/rfc6350#section-6.5.1
        # GEO       https://rfc-editor.org/rfc/rfc6350#section-6.5.2
        # LOGO      https://rfc-editor.org/rfc/rfc6350#section-6.6.3
        # MEMBER    https://rfc-editor.org/rfc/rfc6350#section-6.6.5
        # RELATED   https://rfc-editor.org/rfc/rfc6350#section-6.6.6
        # SOUND     https://rfc-editor.org/rfc/rfc6350#section-6.7.5
        # UID       https://rfc-editor.org/rfc/rfc6350#section-6.7.6
        # URL       https://rfc-editor.org/rfc/rfc6350#section-6.7.8
        # KEY       https://rfc-editor.org/rfc/rfc6350#section-6.8.1
        # FBURL     https://rfc-editor.org/rfc/rfc6350#section-6.9.1
        # CALADRURI https://rfc-editor.org/rfc/rfc6350#section-6.9.2
        # CALURI    https://rfc-editor.org/rfc/rfc6350#section-6.9.3
        elif self._prop.name in (
            "source",
            "photo",
            "tel",
            "impp",
            "geo",
            "logo",
            "member",
            "related",
            "sound",
            "uid",
            "url",
            "key",
            "fburl",
            "caladruri",
            "caluri",
        ):
            self.set_value_with_uri(value, value)

        else:
            self.set_text(value)

    def set_value_with_uri(self, value: str, uri: str) -> None:
        puri = parse_uri(uri)
        if isinstance(uri, InvalidUri):
            self.set_text(value)
            return

        self._uri = puri
        super().set_markup(
            '<a href="{}">{}</a>'.format(  # noqa: UP032
                GLib.markup_escape_text(uri), GLib.markup_escape_text(value)
            )
        )

    def _on_activate_link(self, _label: Gtk.Label, _value: str) -> int:
        assert self._uri is not None
        open_uri(self._uri)
        return Gdk.EVENT_STOP


class SexLabel(Gtk.Label):
    def __init__(self, prop: GenderProperty) -> None:
        Gtk.Label.__init__(self)
        self._prop = prop

        self.set_selectable(True)
        self.set_xalign(0)
        self.set_max_width_chars(50)
        self.set_valign(Gtk.Align.CENTER)
        self.set_halign(Gtk.Align.START)

        self.set_text(prop.sex or "")

    def set_text(self, value: str) -> None:  # type: ignore
        if not value or value == "-":
            super().set_text("")
        else:
            super().set_text(SEX_VALUES[value])


class IdentityLabel(Gtk.Label):
    def __init__(self, prop: GenderProperty) -> None:
        Gtk.Label.__init__(self)
        self._prop = prop

        self.set_selectable(True)
        self.set_xalign(0)
        self.set_max_width_chars(50)
        self.set_valign(Gtk.Align.CENTER)
        self.set_halign(Gtk.Align.START)

        self.set_text(prop.identity or "")

    def set_text(self, value: str) -> None:  # type: ignore
        super().set_text("" if not value else value)


class ValueEntry(Gtk.Entry):
    def __init__(self, prop: TextEntryPropertiesT | BDayProperty) -> None:
        Gtk.Entry.__init__(self)
        self.set_valign(Gtk.Align.CENTER)
        self.set_max_width_chars(50)
        if isinstance(prop, OrgProperty):
            self.set_text(prop.values[0] if prop.values else "")
        else:
            self.set_text(prop.value)


class AdrEntry(Gtk.Entry):
    def __init__(self, prop: AdrProperty, type_: str) -> None:
        Gtk.Entry.__init__(self)
        self.set_valign(Gtk.Align.CENTER)
        self.set_max_width_chars(50)
        values = getattr(prop, type_)
        if not values:
            value = ""
        else:
            value = values[0]
        self.set_text(value)
        self.set_placeholder_text(ADR_PLACEHOLDER_TEXT.get(type_))


class IdentityEntry(Gtk.Entry):
    def __init__(self, prop: GenderProperty):
        Gtk.Entry.__init__(self)
        self.set_valign(Gtk.Align.CENTER)
        self.set_max_width_chars(50)
        self.set_text("" if not prop.identity else prop.identity)


class AdrBox(Gtk.Box, SignalManager):
    __gsignals__ = {
        "field-changed": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (str, str),
        )
    }

    def __init__(self, prop: AdrProperty) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL, spacing=6)
        SignalManager.__init__(self)

        for field in ADR_FIELDS:
            entry = AdrEntry(prop, field)
            self._connect(entry, "notify::text", self._on_text_changed, field)
            self.append(entry)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    def _on_text_changed(self, entry: Gtk.Entry, _param: Any, field: str) -> None:
        self.emit("field-changed", field, entry.get_text())


class AdrLabel(Gtk.Label):
    def __init__(self, prop: AdrProperty, type_: str) -> None:
        Gtk.Label.__init__(self)
        self.set_selectable(True)
        self.set_xalign(0)
        self.set_max_width_chars(50)
        self.set_valign(Gtk.Align.CENTER)
        self.set_halign(Gtk.Align.START)
        values = getattr(prop, type_)
        if not values:
            value = ""
        else:
            value = values[0]
        self.set_text(value)

    def set_text(self, value: str) -> None:  # type: ignore
        self.set_visible(bool(value))
        super().set_text(value)


class AdrBoxReadOnly(Gtk.Box):
    def __init__(self, prop: AdrProperty) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL, spacing=6)

        self._labels: dict[str, AdrLabel] = {}

        for field in ADR_FIELDS:
            label = AdrLabel(prop, field)
            self._labels[field] = label
            self.append(label)

    def set_field(self, field: str, value: str) -> None:
        self._labels[field].set_text(value)


class ValueTextView(Gtk.TextView, SignalManager):
    def __init__(self, prop: NoteProperty | KeyProperty) -> None:
        Gtk.TextView.__init__(self)
        SignalManager.__init__(self)

        self.props.right_margin = 8
        self.props.left_margin = 8
        self.props.top_margin = 8
        self.props.bottom_margin = 8

        self.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
        self.set_hexpand(True)
        self.set_valign(Gtk.Align.FILL)

        self._prop = prop
        self.get_buffer().set_text(prop.value)
        self._connect(self.get_buffer(), "notify::text", self._on_text_changed)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.TextView.do_unroot(self)
        app.check_finalize(self)

    def get_text(self) -> str:
        start_iter, end_iter = self.get_buffer().get_bounds()
        return self.get_buffer().get_text(start_iter, end_iter, False)

    def _on_text_changed(self, _buffer: Gtk.TextBuffer, _param: Any) -> None:
        self._prop.value = self.get_text()


class TypeDropDown(GajimDropDown[str], SignalManager):
    def __init__(self, parameters: Parameters) -> None:
        super().__init__()
        SignalManager.__init__(self)

        self.set_valign(Gtk.Align.CENTER)
        self._parameters = parameters

        self.set_data(
            {
                "-": "-",
                "home": _("Home"),
                "work": _("Work"),
            }
        )

        values = self._parameters.get_types()
        if "home" in values:
            self.select_key("home")

        elif "work" in values:
            self.select_key("work")

        else:
            self.select_key("-")

        self._connect(self, "notify::selected", self._on_selected)

    def do_unroot(self) -> None:
        self._disconnect_all()
        super().do_unroot()

    def _on_selected(
        self, _dropdown: GajimDropDown[str], _param: GObject.ParamSpec
    ) -> None:
        key = self.get_selected_key()
        if key == "-":
            self._parameters.remove_types(["work", "home"])

        elif key == "work":
            self._parameters.add_types(["work"])
            self._parameters.remove_types(["home"])

        elif key == "home":
            self._parameters.add_types(["home"])
            self._parameters.remove_types(["work"])

    def get_text(self) -> str:
        item = self.get_selected_item()
        assert item is not None
        if item.key == "-":
            return ""
        return item.value


class GenderDropDown(GajimDropDown[str]):
    def __init__(self, prop: GenderProperty) -> None:
        super().__init__()
        self.set_valign(Gtk.Align.CENTER)
        self.set_halign(Gtk.Align.START)

        self._prop = prop

        data = {"-": "-"}
        data.update(SEX_VALUES)
        self.set_data(data)

        if not prop.sex:
            self.select_key("-")
        else:
            self.select_key(prop.sex)


class TimezoneLabel(Gtk.Label):
    def __init__(self, prop: TzProperty, show_time: bool = False) -> None:
        Gtk.Label.__init__(self)
        self._prop = prop
        self._show_time = show_time

        self.set_selectable(True)
        self.set_xalign(0)
        self.set_max_width_chars(50)
        self.set_ellipsize(Pango.EllipsizeMode.END)
        self.set_valign(Gtk.Align.CENTER)
        self.set_halign(Gtk.Align.START)

        self.update_text()

    def update_text(self) -> None:
        try:
            tzinfo = zoneinfo.ZoneInfo(self._prop.value)
        except Exception:
            remote_dt_str = ""
        else:
            dt_format = app.settings.get("date_time_format")
            remote_dt_str = dt.datetime.now(tz=tzinfo).strftime(dt_format)

        if self._show_time:
            self.set_visible(bool(remote_dt_str))
            super().set_text(remote_dt_str)
        else:
            data = get_zone_data(self._prop.value)
            super().set_text(self._prop.value if data is None else data.key)


class RemoveButton(Gtk.Button):
    def __init__(self) -> None:
        Gtk.Button.__init__(
            self, halign=Gtk.Align.START, valign=Gtk.Align.CENTER, visible=False
        )
        self.add_css_class("image-button")
        image = Gtk.Image.new_from_icon_name("lucide-trash-symbolic")
        self.set_child(image)


class VCardPropertyGui(SignalManager):
    def __init__(self, prop: SupportedPropertiesT) -> None:
        SignalManager.__init__(self)

        self._prop = prop

        self._second_column: list[Gtk.Widget] = []
        self._third_column: list[Gtk.Widget] = []

        self._desc_label = DescriptionLabel(prop.name)
        self._remove_button = RemoveButton()
        self._connect(self._remove_button, "clicked", self._on_remove_clicked)

        self._edit_widgets: list[Gtk.Widget] = [self._remove_button]
        self._read_widgets: list[Gtk.Widget] = []

        self._type_dropdown: TypeDropDown | None = None
        if prop.name in PROPERTIES_WITH_TYPE:
            self._type_dropdown = TypeDropDown(prop.parameters)
            self._connect(
                self._type_dropdown, "notify::selected", self._on_type_changed
            )
            type_ = self._type_dropdown.get_selected_key()
            assert type_ is not None
            icon_name = self._get_icon_name(type_)
            self._type_image = Gtk.Image.new_from_icon_name(icon_name)
            self._type_image.set_tooltip_text(TYPE_VALUES[type_])

            if prop.name == "adr":
                self._type_image.set_valign(Gtk.Align.START)
                self._type_dropdown.set_valign(Gtk.Align.START)

            self._edit_widgets.append(self._type_dropdown)
            self._read_widgets.append(self._type_image)
            self._second_column.extend([self._type_dropdown, self._type_image])

    def destroy(self) -> None:
        self._disconnect_all()
        if self._type_dropdown is not None:
            self._type_dropdown.run_destroy()
        app.check_finalize(self)

    @staticmethod
    def _get_icon_name(type_: str) -> str | None:
        if type_ == "home":
            return "lucide-house"
        if type_ == "work":
            return "lucide-briefcase"
        return None

    def _on_type_changed(
        self, dropdown: TypeDropDown, _param: GObject.ParamSpec
    ) -> None:
        type_ = dropdown.get_selected_key()
        assert type_ is not None
        icon_name = self._get_icon_name(type_)
        self._type_image.set_from_icon_name(icon_name)
        self._type_image.set_tooltip_text(TYPE_VALUES[type_])

    def _on_remove_clicked(self, button: Gtk.Button) -> None:
        vcard_grid = cast(VCardGrid, button.get_parent())
        vcard_grid.remove_property(self)

    @property
    def row_number(self) -> int:
        grid = cast(Gtk.Grid, self._desc_label.get_parent())
        return grid.query_child(self._desc_label).row  # type: ignore

    def get_base_property(self) -> PropertyT:
        return self._prop

    def set_editable(self, enabled: bool) -> None:
        for widget in self._edit_widgets:
            widget.set_visible(enabled)

        for widget in self._read_widgets:
            widget.set_visible(not enabled)

    def add_to_grid(self, grid: Gtk.Grid, row_number: int) -> None:
        # child, left, top, width, height
        grid.attach(self._desc_label, 0, row_number, 1, 1)

        for widget in self._second_column:
            grid.attach(widget, 1, row_number, 1, 1)

        for widget in self._third_column:
            grid.attach(widget, 2, row_number, 1, 1)

        grid.attach(self._remove_button, 3, row_number, 1, 1)


class TextEntryPropertyGui(VCardPropertyGui):
    def __init__(self, prop: TextEntryPropertiesT, account: str) -> None:
        VCardPropertyGui.__init__(self, prop)

        self._value_entry = ValueEntry(prop)
        self._connect(self._value_entry, "notify::text", self._on_text_changed)

        self._value_label = ValueLabel(prop, account)

        self._edit_widgets.append(self._value_entry)
        self._read_widgets.append(self._value_label)

        self._third_column = [self._value_entry, self._value_label]

    def destroy(self) -> None:
        VCardPropertyGui.destroy(self)
        self._value_label.run_destroy()

    def _on_text_changed(self, entry: Gtk.Entry, _param: Any) -> None:
        text = entry.get_text()
        if isinstance(self._prop, OrgProperty):
            self._prop.values = [text]
        else:
            assert isinstance(self._prop, TextEntryPropertiesT)
            self._prop.value = text
        self._value_label.set_value(text)


class MultiLinePropertyGui(VCardPropertyGui):
    def __init__(self, prop: NoteProperty, _account: str) -> None:
        VCardPropertyGui.__init__(self, prop)

        self._edit_text_view = ValueTextView(prop)
        self._edit_text_view.set_visible(True)

        self._edit_scrolled = Gtk.ScrolledWindow()
        self._edit_scrolled.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self._edit_scrolled.set_child(self._edit_text_view)
        self._edit_scrolled.set_valign(Gtk.Align.CENTER)
        self._edit_scrolled.set_size_request(-1, 100)
        self._edit_scrolled.set_has_frame(True)

        self._read_text_view = ValueTextView(prop)
        self._read_text_view.set_sensitive(False)
        self._read_text_view.set_left_margin(0)
        self._read_text_view.set_visible(True)

        self._read_scrolled = Gtk.ScrolledWindow()
        self._read_scrolled.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self._read_scrolled.set_child(self._read_text_view)
        self._read_scrolled.set_valign(Gtk.Align.CENTER)
        self._read_scrolled.set_size_request(-1, 100)
        self._read_scrolled.add_css_class("profile-scrolled-read")

        self._edit_widgets.append(self._edit_scrolled)
        self._read_widgets.append(self._read_scrolled)

        self._third_column = [self._edit_scrolled, self._read_scrolled]


class DatePropertyGui(VCardPropertyGui):
    def __init__(self, prop: BDayProperty, account: str) -> None:
        VCardPropertyGui.__init__(self, prop)

        self._box = Gtk.Box(spacing=6)
        self._value_entry = ValueEntry(prop)
        self._value_entry.set_placeholder_text(_("YYYY-MM-DD"))
        self._connect(self._value_entry, "notify::text", self._on_text_changed)

        self._calendar_button = Gtk.MenuButton()
        image = Gtk.Image.new_from_icon_name("lucide-calendar-symbolic")
        self._calendar_button.set_child(image)
        self._connect(
            self._calendar_button, "notify::active", self._on_calendar_button_clicked
        )
        self._box.append(self._value_entry)
        self._box.append(self._calendar_button)

        self.calendar = Gtk.Calendar(year=1980, month=5, day=15)
        self.calendar.set_visible(True)
        self._connect(self.calendar, "day-selected", self._on_calendar_day_selected)

        popover = Gtk.Popover()
        popover.set_child(self.calendar)
        self._calendar_button.set_popover(popover)

        self._value_label = ValueLabel(prop, account)

        self._edit_widgets.append(self._box)
        self._read_widgets.append(self._value_label)

        self._third_column = [self._box, self._value_label]

    def destroy(self) -> None:
        VCardPropertyGui.destroy(self)
        self._value_label.run_destroy()

    def _on_text_changed(self, entry: Gtk.Entry, _param: Any) -> None:
        text = entry.get_text()
        assert isinstance(self._prop, BDayProperty)
        self._prop.value = text
        self._value_label.set_value(text)

    def _on_calendar_button_clicked(self, *args: Any) -> None:
        birthday = self._value_entry.get_text()
        if not birthday:
            return
        try:
            date = dt.date.fromisoformat(birthday)
        except ValueError:
            return

        self.calendar.select_day(convert_py_to_glib_datetime(date))

    def _on_calendar_day_selected(self, calendar: Gtk.Calendar) -> None:
        date_time = calendar.get_date()
        date = dt.date(*date_time.get_ymd())
        self._value_entry.set_text(date.isoformat())


class KeyPropertyGui(VCardPropertyGui):
    def __init__(self, prop: KeyProperty, _account: str) -> None:
        VCardPropertyGui.__init__(self, prop)

        self._value_text_view = ValueTextView(prop)
        self._value_text_view.set_visible(True)

        self._scrolled_window = Gtk.ScrolledWindow()
        self._scrolled_window.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self._scrolled_window.set_child(self._value_text_view)
        self._scrolled_window.set_valign(Gtk.Align.CENTER)
        self._scrolled_window.set_size_request(350, 200)
        self._scrolled_window.set_has_frame(True)

        self._copy_button = Gtk.Button.new_from_icon_name("lucide-copy-symbolic")
        self._connect(self._copy_button, "clicked", self._on_copy_clicked)
        self._copy_button.set_halign(Gtk.Align.START)
        self._copy_button.set_valign(Gtk.Align.CENTER)

        self._edit_widgets.append(self._scrolled_window)
        self._read_widgets.append(self._copy_button)

        self._third_column = [self._scrolled_window, self._copy_button]

    def _on_copy_clicked(self, _button: Gtk.Button) -> None:
        app.window.get_clipboard().set(self._value_text_view.get_text())


class GenderPropertyGui(VCardPropertyGui):
    def __init__(self, prop: GenderProperty, _account: str) -> None:
        VCardPropertyGui.__init__(self, prop)

        value_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=12)
        self._value_dropdown = GenderDropDown(prop)
        self._connect(self._value_dropdown, "notify::selected", self._on_selected)

        self._value_entry = IdentityEntry(prop)
        self._value_entry.set_visible(True)
        self._connect(self._value_entry, "notify::text", self._on_text_changed)

        value_box.append(self._value_dropdown)
        value_box.append(self._value_entry)

        label_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=12)
        self._identity_label = IdentityLabel(prop)
        self._identity_label.set_visible(True)
        self._sex_label = SexLabel(prop)
        self._sex_label.set_visible(True)

        label_box.append(self._sex_label)
        label_box.append(self._identity_label)

        self._edit_widgets.append(value_box)
        self._read_widgets.append(label_box)

        self._third_column = [value_box, label_box]

    def destroy(self) -> None:
        VCardPropertyGui.destroy(self)
        self._value_dropdown.run_destroy()

    def _on_text_changed(self, entry: Gtk.Entry, _param: Any) -> None:
        text = entry.get_text()
        assert isinstance(self._prop, GenderProperty)
        self._prop.identity = text
        self._identity_label.set_text(text)

    def _on_selected(self, drowdown: GenderDropDown, _param: Any) -> None:
        sex = drowdown.get_selected_key()
        assert isinstance(self._prop, GenderProperty)
        self._prop.sex = None if sex == "-" else sex
        assert sex is not None
        self._sex_label.set_text(sex)


class AdrPropertyGui(VCardPropertyGui):
    def __init__(self, prop: AdrProperty, _account: str) -> None:
        VCardPropertyGui.__init__(self, prop)

        self._entry_box = AdrBox(prop)
        self._connect(self._entry_box, "field-changed", self._on_field_changed)

        self._read_box = AdrBoxReadOnly(prop)

        self._edit_widgets.append(self._entry_box)
        self._read_widgets.append(self._read_box)

        self._third_column = [self._entry_box, self._read_box]

    def _on_field_changed(self, _box: Gtk.Box, field: str, value: str) -> None:
        setattr(self._prop, field, [value])
        self._read_box.set_field(field, value)


class TzPropertyGui(VCardPropertyGui):
    def __init__(self, prop: TzProperty, account: str) -> None:
        VCardPropertyGui.__init__(self, prop)

        self._value_dropdown: GajimDropDown[str] = GajimDropDown(data=ZONES)
        self._value_dropdown.set_enable_search(True)
        self._value_dropdown.select_key(prop.value)

        self._connect(
            self._value_dropdown,
            "notify::selected",
            self._on_zone_selected,
        )

        self._read_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        self._value_label = TimezoneLabel(prop)
        self._time_label = TimezoneLabel(prop, show_time=True)

        self._read_box.append(self._value_label)
        self._read_box.append(self._time_label)

        self._edit_widgets.append(self._value_dropdown)
        self._read_widgets.append(self._read_box)

        self._third_column = [self._value_dropdown, self._read_box]

    def destroy(self) -> None:
        VCardPropertyGui.destroy(self)
        self._value_dropdown.run_destroy()

    def _on_zone_selected(self, dropdown: GajimDropDown[str], *args: Any) -> None:
        item = dropdown.get_selected_item()
        assert item is not None
        assert isinstance(self._prop, TzProperty)
        self._prop.value = item.key
        self._value_label.update_text()
        self._time_label.update_text()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import cast
from typing import NamedTuple

import logging
from enum import IntEnum

from gi.repository import Gtk
from nbxmpp.errors import StanzaError
from nbxmpp.structs import AffiliationResult
from nbxmpp.task import Task

from gajim.common import app
from gajim.common.client import Client
from gajim.common.i18n import _
from gajim.common.modules.contacts import GroupchatContact

from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.apply_button_box import ApplyButtonBox
from gajim.gtk.builder import get_builder
from gajim.gtk.util.classes import SignalManager

log = logging.getLogger("gajim.gtk.groupchat_affiliation")


class AffiliationRow(NamedTuple):
    jid: str
    nick: str | None
    affiliation: str


class Column(IntEnum):
    JID = 0
    NICK = 1
    AFFILIATION = 2
    AFFILIATION_TEXT = 3


class GroupchatAffiliation(Gtk.Box, SignalManager):
    def __init__(self, client: Client, contact: GroupchatContact) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)
        SignalManager.__init__(self)

        self._client = client
        self._contact = contact
        self._current_rows: dict[str, AffiliationRow] = {}

        self_contact = contact.get_self()
        assert self_contact is not None
        self._own_affiliation = self_contact.affiliation.value

        self._ui = get_builder("groupchat_affiliation.ui")

        self._connect(
            self._ui.affiliation_selection, "changed", self._on_selection_changed
        )
        self._connect(self._ui.address_renderer, "edited", self._on_jid_edited)
        self._connect(self._ui.reserved_name_renderer, "edited", self._on_nick_edited)
        self._connect(
            self._ui.affiliation_renderer, "changed", self._on_affiliation_changed
        )
        self._connect(self._ui.add_button, "clicked", self._on_add)
        self._connect(self._ui.remove_button, "clicked", self._on_remove)

        self._treeview = self._ui.affiliation_treeview
        self._store = self._ui.affiliation_store

        self.append(self._ui.main_box)

        self._apply_button = ApplyButtonBox(_("Apply"), on_clicked=self._on_apply)
        self._apply_button.set_hexpand(True)
        self._apply_button.set_halign(Gtk.Align.END)
        self._ui.button_box.append(self._apply_button)

        if self._own_affiliation in ("admin", "owner"):
            self._ui.add_button.set_sensitive(True)
            self._ui.add_button.set_tooltip_text("")

        for affiliation in ("owner", "admin", "member"):
            self._client.get_module("MUC").get_affiliation(
                self._contact.jid,
                affiliation,
                callback=self._on_affiliations_received,
                user_data=affiliation,
            )

    def do_unroot(self) -> None:
        self._disconnect_all()
        del self._treeview
        del self._store
        del self._apply_button
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    def _on_apply(self, _button: Gtk.Button) -> None:
        self._begin_progress()
        self._set_affiliations()

    def _begin_progress(self) -> None:
        self._ui.affiliation_scrolled.set_sensitive(False)
        self._ui.add_remove_button_box.set_sensitive(False)

    def _end_progress(self) -> None:
        self._ui.affiliation_scrolled.set_sensitive(True)
        self._ui.add_remove_button_box.set_sensitive(True)

    def _on_add(self, _button: Gtk.Button) -> None:
        affiliation_edit, jid_edit = self._allowed_to_edit("member")

        iter_ = self._store.append(
            [None, None, "member", _("Member"), affiliation_edit, jid_edit]
        )

        path = self._store.get_path(iter_)
        self._treeview.scroll_to_cell(path, None, False, 0, 0)
        self._treeview.get_selection().unselect_all()
        self._treeview.get_selection().select_path(path)

        self._update_apply_button_state()

    def _on_remove(self, _button: Gtk.Button) -> None:
        selection = self._treeview.get_selection()
        rows = selection.get_selected_rows()
        if rows is None:
            return

        _model, paths = rows

        owner_count = len(
            [row for row in self._store if row[Column.AFFILIATION] == "owner"]
        )
        references: list[Gtk.TreeRowReference] = []
        for path in paths:
            if self._store[path][Column.AFFILIATION] == "owner":
                if owner_count < 2:
                    # There must be at least one owner
                    InformationAlertDialog(
                        _("Error"), _("A Group Chat needs at least one Owner")
                    )
                    return
                owner_count -= 1
            references.append(Gtk.TreeRowReference(self._store, path))

        for ref in references:
            path = ref.get_path()
            assert path is not None
            iter_ = self._store.get_iter(path)
            self._store.remove(iter_)

        self._update_apply_button_state()

    def _on_jid_edited(
        self, _renderer: Gtk.CellRendererText, path: str, new_text: str
    ) -> None:
        old_text = self._store[path][Column.JID]
        if new_text == old_text:
            return

        if self._jid_exists(new_text):
            self._raise_error()
            return

        self._store[path][Column.JID] = new_text
        self._update_apply_button_state()

    def _on_nick_edited(
        self, _renderer: Gtk.CellRendererText, path: str, new_text: str
    ) -> None:
        self._store[path][Column.NICK] = new_text or None
        self._update_apply_button_state()

    def _on_affiliation_changed(
        self,
        cell_renderer_combo: Gtk.CellRendererCombo,
        path_string: str,
        new_iter: Gtk.TreeIter,
    ) -> None:
        combo_store = cell_renderer_combo.get_property("model")
        affiliation_text = combo_store.get_value(new_iter, 0)
        affiliation = combo_store.get_value(new_iter, 1)

        self._store[path_string][Column.AFFILIATION] = affiliation
        self._store[path_string][Column.AFFILIATION_TEXT] = affiliation_text
        self._update_apply_button_state()

    def _on_selection_changed(self, tree_selection: Gtk.TreeSelection) -> None:
        sensitive = bool(tree_selection.count_selected_rows())
        selected_affiliations = self._get_selected_affiliations(tree_selection)
        self._set_remove_button_state(sensitive, selected_affiliations)

    def _jid_exists(self, jid: str) -> bool:
        return any(row[Column.JID] == jid for row in self._store)

    @staticmethod
    def _get_selected_affiliations(tree_selection: Gtk.TreeSelection) -> set[str]:
        rows = tree_selection.get_selected_rows()
        assert rows is not None
        model, paths = rows
        selected_affiliations: set[str] = set()
        for path in paths:
            selected_affiliations.add(model[path][Column.AFFILIATION])
        return selected_affiliations

    def _set_remove_button_state(
        self, sensitive: bool, selected_affiliations: set[str]
    ) -> None:
        if self._own_affiliation not in ("admin", "owner"):
            self._ui.remove_button.set_sensitive(False)
            return

        self._ui.remove_button.set_tooltip_text("")

        if not sensitive:
            self._ui.remove_button.set_sensitive(False)
            return

        if self._own_affiliation == "owner":
            self._ui.remove_button.set_sensitive(True)
            return

        if {"owner", "admin"}.intersection(selected_affiliations):
            self._ui.remove_button.set_sensitive(False)
            self._ui.remove_button.set_tooltip_text(
                _("You are not allowed to modify the affiliation of Admins and Owners")
            )
            return

        self._ui.remove_button.set_sensitive(True)

    def _update_apply_button_state(self):
        if self._own_affiliation not in ("admin", "owner"):
            return
        new_rows = self._get_new_rows()
        changed = set(new_rows.values()) != set(self._current_rows.values())
        self._apply_button.set_button_state(changed)

    def _allowed_to_edit(self, affiliation: str) -> tuple[bool, bool]:
        if self._own_affiliation == "owner":
            return True, True

        if self._own_affiliation == "admin":
            if affiliation in ("admin", "owner"):
                return False, False
            return False, True
        return False, False

    def _get_new_rows(self) -> dict[str, AffiliationRow]:
        rows: dict[str, AffiliationRow] = {}

        for row in self._store:
            if not row[Column.JID]:
                continue

            rows[row[Column.JID]] = AffiliationRow(
                jid=row[Column.JID],
                nick=row[Column.NICK],
                affiliation=row[Column.AFFILIATION],
            )

        return rows

    def _get_diff(self) -> list[AffiliationRow]:
        new_rows = self._get_new_rows()

        before = set(self._current_rows.keys())
        after = set(new_rows.keys())

        removed = before - after
        removed_rows = [
            row for row in self._current_rows.values() if row.jid in removed
        ]
        removed_rows = [row._replace(affiliation="none") for row in removed_rows]

        added = after - before
        added_rows = [row for jid, row in new_rows.items() if jid in added]

        same = after - removed - added
        same_rows = {row for jid, row in new_rows.items() if jid in same}
        changed_rows = list(same_rows - set(self._current_rows.values()))

        # Find which nickname reservation have been unset and replace the
        # value with an empty string. Only the empty string tells the server
        # to unset the nickname reservation.
        modified_rows: list[AffiliationRow] = []
        for changed in changed_rows:
            before = self._current_rows[changed.jid]
            if changed.nick is None and before.nick is not None:
                # Nick was unset
                modified_rows.append(changed._replace(nick=""))
                continue

            modified_rows.append(changed)

        return removed_rows + added_rows + modified_rows

    def _set_affiliations(self) -> None:
        diff_rows = self._get_diff()

        affiliations: dict[str, dict[str, str]] = {}
        for row in diff_rows:
            affiliations[row.jid] = {"affiliation": row.affiliation}
            if row.nick is not None:
                affiliations[row.jid]["nick"] = row.nick

        for jid, data in affiliations.items():
            # Send only one change at a time to work around a old prosody bug
            # https://issues.prosody.im/345
            self._client.get_module("MUC").set_affiliation(
                self._contact.jid, {jid: data}, callback=self._on_affiliation_finished
            )

    def _on_affiliation_finished(self, task: Task) -> None:
        try:
            task.finish()
        except StanzaError as error:
            log.info("Error while setting affiliations: %s", error)
            self._end_progress()
            self._apply_button.set_error(str(error))
            return

        self._current_rows = self._get_new_rows()
        self._end_progress()
        self._apply_button.set_success()

    def _on_affiliations_received(self, task: Task) -> None:
        affiliation = task.get_user_data()
        try:
            result = cast(AffiliationResult, task.finish())
        except StanzaError as error:
            log.info(
                "Error while requesting %s affiliations: %s",
                affiliation,
                error.condition,
            )
            return

        for jid, attrs in result.users.items():
            affiliation_edit, jid_edit = self._allowed_to_edit(affiliation)
            nick = attrs.get("nick") or None

            jid = str(jid)
            self._store.append(
                [
                    jid,
                    nick,
                    affiliation,
                    _(affiliation.capitalize()),
                    affiliation_edit,
                    jid_edit,
                ]
            )

            self._current_rows[jid] = AffiliationRow(
                jid=jid, nick=nick, affiliation=affiliation
            )

    @staticmethod
    def _raise_error() -> None:
        InformationAlertDialog(
            _("Error"), _("An entry with this XMPP Address already exists")
        )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import Gtk

from gajim.common.const import AvatarSize
from gajim.common.i18n import _
from gajim.common.util.datetime import utc_now
from gajim.common.util.status import get_uf_show
from gajim.common.util.text import process_non_spacing_marks

from gajim.gtk.avatar import get_show_circle
from gajim.gtk.conversation.message_widget import MessageWidget
from gajim.gtk.conversation.rows.base import BaseRow
from gajim.gtk.conversation.rows.widgets import DateTimeLabel
from gajim.gtk.conversation.rows.widgets import SimpleLabel
from gajim.gtk.util.misc import convert_surface_to_texture


class UserStatus(BaseRow):
    def __init__(self, account: str, name: str, show: str, status: str | None) -> None:
        BaseRow.__init__(self, account)

        self.type = "muc-user-status"
        timestamp = utc_now()
        self.timestamp = timestamp.astimezone()
        self.db_timestamp = timestamp.timestamp()

        avatar_placeholder = Gtk.Box()
        avatar_placeholder.set_size_request(AvatarSize.ROSTER, -1)
        self.grid.attach(avatar_placeholder, 0, 0, 1, 1)

        show_icon = Gtk.Image(pixel_size=AvatarSize.SHOW_CIRCLE)
        show_icon.set_opacity(0.6)
        surface = get_show_circle(show, 16, self.get_scale_factor())
        show_icon.set_from_paintable(convert_surface_to_texture(surface))
        self.grid.attach(show_icon, 1, 0, 1, 1)

        show = get_uf_show(show)

        message = _("{nick} is now {show}").format(nick=name, show=show)

        self._label = SimpleLabel()
        self._label.set_valign(Gtk.Align.END)
        self._label.set_text(process_non_spacing_marks(message))
        self._label.add_css_class("gajim-status-message")
        self.grid.attach(self._label, 2, 0, 1, 1)

        if status is not None:
            message_widget = MessageWidget(account)
            message_widget.add_css_class("gajim-status-message")
            message_widget.add_with_styling(status)
            self.grid.attach(message_widget, 2, 1, 1, 1)

        timestamp_widget = DateTimeLabel(self.timestamp)
        timestamp_widget.set_halign(Gtk.Align.START)
        timestamp_widget.set_valign(Gtk.Align.END)
        self.grid.attach(timestamp_widget, 3, 0, 1, 1)

    def do_unroot(self) -> None:
        BaseRow.do_unroot(self)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0153: vCard-Based Avatars

from __future__ import annotations

from typing import Any

import weakref
from collections.abc import Callable
from collections.abc import Generator

from nbxmpp.const import AvatarState
from nbxmpp.errors import is_error
from nbxmpp.modules.vcard_temp import VCard
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.protocol import Presence
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import PresenceProperties
from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common import types
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.util import as_task
from gajim.common.task_manager import Task
from gajim.common.util.classes import TTLCache

VCardContactsT = BareContact | GroupchatContact | GroupchatParticipant
NS_AVATAR_HASH = "muc#roominfo_avatarhash"
NS_AVATAR_HASH_TEMP = "{http://modules.prosody.im/mod_vcard_muc}avatar#sha1"


class VCardAvatars(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)
        self._ignored_sha_cache: TTLCache[tuple[JID, str], None] = TTLCache(
            ttl_seconds=60 * 60 * 6, extend_ttl_on_hit=False
        )
        self._muc_avatar_cache: dict[JID, str] = {}
        self.avatar_conversion_available = False

        self.handlers = [
            StanzaHandler(
                name="presence",
                typ="available",
                callback=self._presence_received,
                ns=Namespace.VCARD_UPDATE,
                priority=51,
            ),
        ]

    def pass_disco(self, info: DiscoInfo) -> None:
        is_available = Namespace.VCARD_CONVERSION in info.features
        self.avatar_conversion_available = is_available
        self._log.info("Discovered Avatar Conversion")

    def get_avatar_sha(self, jid: JID) -> str | None:
        return self._muc_avatar_cache.get(jid)

    def invalidate_cache(self, jid: JID) -> None:
        self._muc_avatar_cache.pop(jid, None)

    @as_task
    def _request_vcard(
        self, contact: types.ChatContactT, expected_sha: str
    ) -> Generator[Any, Any]:
        _task = yield

        if app.app.avatar_storage.avatar_exists(expected_sha):
            # Check if avatar was already received while task was queued
            contact.update_avatar(expected_sha)
            self._log.info("Found avatar in storage: %s %s", contact.jid, expected_sha)
            return

        vcard = yield self._con.get_module("VCardTemp").request_vcard(jid=contact.jid)

        cache_key = (contact.jid, expected_sha)

        if is_error(vcard):
            self._log.warning(vcard)
            self._ignored_sha_cache.add(cache_key, None)
            return

        assert isinstance(vcard, VCard)
        avatar, avatar_sha = vcard.get_avatar()
        if avatar is None or avatar_sha is None:
            self._log.info("Avatar missing: %s %s", contact.jid, expected_sha)
            self._ignored_sha_cache.add(cache_key, None)
            return

        if expected_sha != avatar_sha:
            self._log.warning(
                "Avatar mismatch: %s %s != %s", contact.jid, expected_sha, avatar_sha
            )
            self._ignored_sha_cache.add(cache_key, None)
            return

        self._log.info("Received: %s %s", contact.jid, avatar_sha)
        app.app.avatar_storage.save_avatar(avatar)

        if isinstance(contact, BareContact | GroupchatContact):
            app.storage.archive.set_contact_value(
                self._account, contact.jid, "avatar_sha", avatar_sha
            )

        else:
            self._muc_avatar_cache[contact.jid] = avatar_sha

        contact.update_avatar(avatar_sha)

    def _presence_received(
        self,
        _con: types.NBXMPPClient,
        _stanza: Presence,
        properties: PresenceProperties,
    ) -> None:

        if properties.avatar_state in (AvatarState.IGNORE, AvatarState.NOT_READY):
            return

        assert properties.jid is not None
        if self._con.get_own_jid().bare_match(properties.jid):
            return

        if properties.from_muc:
            self._muc_update_received(properties)

        else:
            jid = properties.jid.new_as_bare()
            muc = self._con.get_module("MUC").get_muc_data(properties.jid)
            self._process_update(
                jid, properties.avatar_state, properties.avatar_sha, muc is not None
            )

    def muc_disco_info_update(self, disco_info: DiscoInfo) -> None:
        if not disco_info.supports(Namespace.VCARD):
            return

        ns_avatar_hash = None
        if disco_info.has_field(Namespace.MUC_INFO, NS_AVATAR_HASH):
            ns_avatar_hash = NS_AVATAR_HASH

        elif disco_info.has_field(Namespace.MUC_INFO, NS_AVATAR_HASH_TEMP):
            ns_avatar_hash = NS_AVATAR_HASH_TEMP

        if ns_avatar_hash is None:
            # Workaround so we don’t delete the avatar for servers that don’t
            # support sha in disco info.
            return

        avatar_sha = (
            disco_info.get_field_value(Namespace.MUC_INFO, ns_avatar_hash) or None
        )
        state = AvatarState.EMPTY if not avatar_sha else AvatarState.ADVERTISED

        assert disco_info.jid is not None
        self._process_update(disco_info.jid, state, avatar_sha, True)

    def _process_update(
        self, jid: JID, state: AvatarState, avatar_sha: str | None, groupchat: bool
    ) -> None:
        contact = self._con.get_module("Contacts").get_contact(jid, groupchat=groupchat)
        assert isinstance(contact, BareContact | GroupchatContact)

        if state == AvatarState.EMPTY:
            # Empty <photo/> tag, means no avatar is advertised
            assert avatar_sha is None
            self._log.info("%s has no avatar published", jid)
            app.storage.archive.set_contact_value(
                self._account, contact.jid, "avatar_sha", None
            )
            contact.update_avatar(avatar_sha)

        else:
            assert avatar_sha
            self._log.info("Update: %s %s", jid, avatar_sha)

            if avatar_sha == contact.avatar_sha:
                self._log.info("Avatar already known: %s %s", jid, avatar_sha)
                return

            if app.app.avatar_storage.avatar_exists(avatar_sha):
                # Check if the avatar is already in storage
                self._log.info("Found avatar in storage")
                app.storage.archive.set_contact_value(
                    self._account, contact.jid, "avatar_sha", avatar_sha
                )
                contact.update_avatar(avatar_sha)
                return

            if (jid, avatar_sha) in self._ignored_sha_cache:
                self._log.info(
                    "Avatar will be ignored because of cached error: %s %s",
                    jid,
                    avatar_sha,
                )
                return

            task = VCardAvatarsTask(contact, avatar_sha, self._request_vcard)
            app.task_manager.add_task(task)

    def _muc_update_received(self, properties: PresenceProperties) -> None:
        assert properties.jid is not None
        jid = properties.jid
        avatar_sha = properties.avatar_sha

        contact = self._con.get_module("Contacts").get_contact(
            properties.jid, groupchat=True
        )
        assert isinstance(contact, GroupchatParticipant)

        if contact.is_blocked:
            self._log.info("Ignore avatar because contact is blocked: %s", contact.jid)
            return

        # Custom ejabberd room config option
        if contact.room.get_disco() is not None:
            # We have no disco when we create a new room, but receive our
            # self presence
            allow_query = contact.room.get_config_value(
                "muc#roomconfig_allow_query_users"
            )
            if allow_query is False:
                self._log.debug("Room does not allow IQ queries: %s", contact.jid)
                return

        if properties.avatar_state == AvatarState.EMPTY:
            # Empty <photo/> tag, means no avatar is advertised
            self._log.info("%s has no avatar published", jid)
            self._muc_avatar_cache.pop(jid, None)
            contact.update_avatar()

        else:
            assert avatar_sha
            self._log.info("Update: %s %s", jid, avatar_sha)
            if not app.app.avatar_storage.avatar_exists(avatar_sha):
                if (jid, avatar_sha) in self._ignored_sha_cache:
                    self._log.info(
                        "Avatar will be ignored because of cached error: %s %s",
                        jid,
                        avatar_sha,
                    )
                    return

                app.log("avatar").info("Request avatar: %s", jid)
                task = VCardAvatarsTask(contact, avatar_sha, self._request_vcard)
                app.task_manager.add_task(task)
                return

            current_avatar_sha = self._muc_avatar_cache.get(jid)
            if current_avatar_sha != avatar_sha:
                self._log.info("%s changed their avatar: %s", jid, avatar_sha)
                self._muc_avatar_cache[jid] = avatar_sha
                contact.update_avatar()

            else:
                self._log.info("Avatar already known: %s", jid)


class VCardAvatarsTask(Task):
    def __init__(
        self, contact: VCardContactsT, sha: str, callback: Callable[..., Any]
    ) -> None:

        Task.__init__(self)
        self._contact = contact
        self._sha = sha
        self._callback = weakref.WeakMethod(callback)

    def execute(self) -> None:
        callback = self._callback()
        if callback is not None:
            callback(self._contact, self._sha)

    def preconditions_met(self) -> bool:
        try:
            client = app.get_client(self._contact.account)
        except Exception:
            self.set_obsolete()
            return False

        if not client.is_available():
            self.set_obsolete()
            return False

        if isinstance(self._contact, GroupchatParticipant):
            if not self._contact.room.is_joined:
                if self._contact.room.is_not_joined:
                    self.set_obsolete()
                return False

        return True

    def __repr__(self) -> str:
        return f"VCardAvatars ({self._contact.jid} {self._sha})"

    def __hash__(self) -> int:
        return hash((self._contact, self._sha))



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0085: Chat State Notifications

from __future__ import annotations

from typing import Any

import time
from collections import defaultdict
from functools import wraps
from itertools import chain

from gi.repository import GLib
from nbxmpp.const import Chatstate as State
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.protocol import NodeProcessed
from nbxmpp.protocol import Presence
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import PresenceProperties
from nbxmpp.structs import StanzaHandler

from gajim.common import types
from gajim.common.const import ClientState
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.structs import OutgoingMessage

INACTIVE_AFTER = 60
PAUSED_AFTER = 10
REMOTE_PAUSED_AFTER = 30


def ensure_enabled(func: Any) -> Any:
    @wraps(func)
    def func_wrapper(self: Any, *args: Any, **kwargs: Any):
        # pylint: disable=protected-access
        if not self._enabled:
            return None
        return func(self, *args, **kwargs)

    return func_wrapper


class Chatstate(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(
                name="presence",
                callback=self._presence_received,
                typ="error",
                priority=50,
            ),
            StanzaHandler(
                name="presence",
                callback=self._presence_received,
                typ="unavailable",
                priority=50,
            ),
            StanzaHandler(
                name="message",
                typ="chat",
                callback=self._process_chatstate,
                ns=Namespace.CHATSTATES,
                priority=46,
            ),
            StanzaHandler(
                name="message",
                typ="groupchat",
                callback=self._process_groupchat_chatstate,
                ns=Namespace.CHATSTATES,
                priority=46,
            ),
        ]

        # Our current chatstate with a specific contact
        self._chatstates: dict[JID, State] = {}

        # The current chatstate we received from a contact
        self._remote_chatstate: dict[JID, State] = {}
        # Cache set of participants that are composing for group chats,
        # to avoid having to iterate over all their chat states to determine
        # who is typing a message.
        self._muc_composers: dict[JID, set[JID]] = defaultdict(set)

        self._remote_composing_timeouts: dict[tuple[JID, str], int] = {}

        self._last_keyboard_activity: dict[JID, float] = {}
        self._last_mouse_activity: dict[JID, float] = {}
        self._timeout_id = None
        self._delay_timeout_ids: dict[JID, int] = {}
        self._blocked: list[JID] = []
        self._enabled = False

        self._client.connect_signal("state-changed", self._on_client_state_changed)
        self._client.connect_signal("resume-failed", self._on_client_resume_failed)

    def _on_client_resume_failed(
        self, _client: types.Client, _signal_name: str
    ) -> None:
        self._set_enabled(False)

    def _on_client_state_changed(
        self, _client: types.Client, _signal_name: str, state: ClientState
    ) -> None:
        if state.is_disconnected:
            self._set_enabled(False)
        elif state.is_connected:
            self._set_enabled(True)

    def _set_enabled(self, value: bool) -> None:
        if self._enabled == value:
            return

        self._log.info("Chatstate module %s", "enabled" if value else "disabled")
        self._enabled = value

        if value:
            self._timeout_id = GLib.timeout_add_seconds(2, self._check_last_interaction)
        else:
            self.cleanup()
            self._client.get_module("Contacts").force_chatstate_update()

    @ensure_enabled
    def _presence_received(
        self,
        _con: types.NBXMPPClient,
        _stanza: Presence,
        properties: PresenceProperties,
    ) -> None:
        if properties.is_self_bare:
            return

        jid = properties.jid

        assert jid is not None
        self._remote_chatstate.pop(jid, None)
        self._chatstates.pop(jid, None)
        self._last_mouse_activity.pop(jid, None)
        self._last_keyboard_activity.pop(jid, None)

        self._log.info("Reset chatstate for %s", jid)

        contact = self._get_contact_if_exists(jid)
        if contact is None:
            return

        contact.notify("chatstate-update")

    def _raise_if_necessary(self, properties: MessageProperties) -> None:
        if properties.chatstate != State.ACTIVE:
            raise NodeProcessed

    def _process_chatstate(
        self, _con: types.NBXMPPClient, _stanza: Any, properties: MessageProperties
    ) -> None:
        if properties.chatstate is None:
            return

        if (
            properties.is_self_message
            or properties.is_mam_message
            or (properties.carbon is not None and properties.carbon.is_sent)
        ):
            return self._raise_if_necessary(properties)

        jid = properties.remote_jid
        assert jid is not None

        m_type = "chat"
        state = properties.chatstate
        self._remote_chatstate[jid] = state

        self._log.info("Recv: %-10s - %s (%s)", state, jid, m_type)

        self._set_composing_timeout(jid, m_type, state)

        contact = self._get_contact_if_exists(jid)
        if contact is not None:
            contact.notify("chatstate-update")

        return self._raise_if_necessary(properties)

    def _process_groupchat_chatstate(
        self, _con: types.NBXMPPClient, _stanza: Any, properties: MessageProperties
    ) -> None:
        if properties.chatstate is None:
            return

        jid = properties.jid
        assert jid is not None

        if properties.is_mam_message or jid.is_bare:
            return self._raise_if_necessary(properties)

        muc_jid = jid.new_as_bare()

        data = self._con.get_module("MUC").get_muc_data(muc_jid)
        if data is None or data.nick == jid.resource:
            # Chatstate from our own joined jid resource
            return self._raise_if_necessary(properties)

        m_type = "groupchat"
        state = properties.chatstate
        self._log.info("Recv: %-10s - %s (%s)", state, jid, m_type)

        self._set_composing_timeout(jid, m_type, state)

        if state == State.COMPOSING:
            self._muc_composers[muc_jid].add(jid)
        else:
            self._muc_composers[muc_jid].discard(jid)

        contact = self._get_contact_if_exists(muc_jid)
        if contact is not None:
            contact.notify("chatstate-update")

        self._raise_if_necessary(properties)

    def _set_composing_timeout(self, jid: JID, m_type: str, state: State) -> None:
        self._remove_remote_composing_timeout(jid, m_type)
        if state not in (State.COMPOSING, State.PAUSED):
            return

        # the spec does not cover any timeout for the composing action,
        # but if a contact's client does not send another chat state,
        # we don't want the GUI to show that they are "composing" forever
        self._remote_composing_timeouts[(jid, m_type)] = GLib.timeout_add_seconds(
            REMOTE_PAUSED_AFTER, self._on_remote_composing_timeout, jid, m_type
        )

    def _on_remote_composing_timeout(self, jid: JID, m_type: str) -> None:
        self._remote_composing_timeouts.pop((jid, m_type), None)
        self._log.info(
            "Set to ACTIVE after timeout has been reached - %s (%s)", jid, m_type
        )

        if m_type == "groupchat":
            muc_jid = jid.new_as_bare()
            self._muc_composers[muc_jid].discard(jid)
            contact = self._get_contact_if_exists(muc_jid)

        else:
            self._remote_chatstate[jid] = State.ACTIVE
            contact = self._get_contact_if_exists(jid)

        if contact is not None:
            contact.notify("chatstate-update")

    def get_composers(self, muc_jid: JID) -> list[GroupchatParticipant]:
        """
        List of group chat participants that are composing (=typing) for a MUC.
        """
        composers: list[GroupchatParticipant] = []
        for jid in self._muc_composers[muc_jid]:
            contact = self._get_contact(jid, groupchat=True)
            assert isinstance(contact, GroupchatParticipant)
            composers.append(contact)

        return composers

    def _remove_remote_composing_timeout(self, jid: JID, m_type: str):
        source_id = self._remote_composing_timeouts.pop((jid, m_type), None)
        if source_id is not None:
            self._log.debug("Removing remote composing timeout of %s (%s)", jid, m_type)
            GLib.source_remove(source_id)

    @ensure_enabled
    def _check_last_interaction(self) -> bool:
        now = time.time()
        for jid in list(self._last_mouse_activity.keys()):
            time_ = self._last_mouse_activity[jid]
            current_state = self._chatstates.get(jid)
            if current_state is None:
                self._last_mouse_activity.pop(jid, None)
                self._last_keyboard_activity.pop(jid, None)
                continue

            if current_state in (State.GONE, State.INACTIVE):
                continue

            new_chatstate = None
            if now - time_ > INACTIVE_AFTER:
                new_chatstate = State.INACTIVE

            elif current_state == State.COMPOSING:
                key_time = self._last_keyboard_activity[jid]
                if now - key_time > PAUSED_AFTER:
                    new_chatstate = State.PAUSED

            if new_chatstate is not None:
                if self._chatstates.get(jid) != new_chatstate:
                    contact = self._get_contact_if_exists(jid)
                    if contact is not None:
                        self.set_chatstate(contact, new_chatstate)

        return GLib.SOURCE_CONTINUE

    def get_remote_chatstate(self, jid: JID) -> State | None:
        return self._remote_chatstate.get(jid)

    @ensure_enabled
    def set_active(self, contact: types.ChatContactT) -> None:
        if contact.settings.get("send_chatstate") == "disabled":
            return
        self._last_mouse_activity[contact.jid] = time.time()
        self._chatstates[contact.jid] = State.ACTIVE

    def get_active_chatstate(self, contact: types.ChatContactT) -> str | None:
        # determines if we add 'active' on outgoing messages
        if contact.settings.get("send_chatstate") == "disabled":
            return None

        if not contact.is_groupchat:
            # Don’t send chatstates to ourself
            if self._client.is_own_jid(contact.jid):
                return None

            if not contact.supports(Namespace.CHATSTATES):
                return None

        self.set_active(contact)
        return "active"

    @ensure_enabled
    def block_chatstates(self, contact: types.ChatContactT, block: bool) -> None:
        # Block sending chatstates to a contact
        # Used for example if we cycle through the MUC nick list, which
        # produces a lot of buffer 'changed' signals from the input textview.
        # This would lead to sending ACTIVE -> COMPOSING -> ACTIVE ...
        if block:
            self._blocked.append(contact.jid)
        else:
            self._blocked.remove(contact.jid)

    @ensure_enabled
    def set_chatstate_delayed(self, contact: types.ChatContactT, state: State) -> None:
        # Used when we go from Composing -> Active after deleting all text
        # from the Textview. We delay the Active state because maybe the
        # User starts writing again.

        # Don’t send chatstates to ourself
        if self._client.is_own_jid(contact.jid):
            return

        self._remove_delay_timeout(contact)
        self._delay_timeout_ids[contact.jid] = GLib.timeout_add_seconds(
            2, self.set_chatstate, contact, state
        )

    @staticmethod
    def _get_chatstate_setting(contact: types.ChatContactT) -> str:
        if isinstance(contact, GroupchatParticipant):
            return contact.room.settings.get("send_chatstate")
        return contact.settings.get("send_chatstate")

    @ensure_enabled
    def set_chatstate(self, contact: types.ChatContactT, state: State) -> None:
        # Don’t send chatstates to ourself
        if self._client.is_own_jid(contact.jid):
            return

        if contact.jid in self._blocked:
            return

        self._remove_delay_timeout(contact)
        current_state = self._chatstates.get(contact.jid)
        setting = self._get_chatstate_setting(contact)
        if setting == "disabled":
            # Send a last 'active' state after user disabled chatstates
            if current_state is not None:
                self._log.info("Disabled for %s", contact)
                self._log.info("Send last state: %-10s - %s", State.ACTIVE, contact)

                self._send_chatstate(contact, State.ACTIVE)

            self._chatstates.pop(contact.jid, None)
            self._last_mouse_activity.pop(contact.jid, None)
            self._last_keyboard_activity.pop(contact.jid, None)
            return

        if isinstance(contact, BareContact):
            if not contact.is_subscribed:
                self._log.debug("Contact not subscribed: %s", contact)
                return

            if not contact.is_available:
                self._log.debug("Contact offline: %s", contact)
                return

        elif isinstance(contact, GroupchatParticipant):
            if not contact.is_available:
                self._log.debug("Contact offline: %s", contact)
                return

        else:
            if not contact.is_joined:
                self._log.debug("Groupchat not joined: %s", contact)
                return

        if state in (State.ACTIVE, State.COMPOSING):
            self._last_mouse_activity[contact.jid] = time.time()

        if setting == "composing_only":
            if state in (State.INACTIVE, State.GONE):
                state = State.ACTIVE

        if current_state == state:
            return

        self._log.info("Send: %-10s - %s", state, contact)

        self._send_chatstate(contact, state)

        self._chatstates[contact.jid] = state

    def _send_chatstate(self, contact: types.ChatContactT, chatstate: State) -> None:
        message = OutgoingMessage(
            account=self._account,
            contact=contact,
            chatstate=chatstate.value,
            play_sound=False,
        )

        self._client.send_message(message)

    @ensure_enabled
    def set_mouse_activity(self, contact: types.ChatContactT, was_paused: bool) -> None:
        if contact.settings.get("send_chatstate") == "disabled":
            return
        self._last_mouse_activity[contact.jid] = time.time()
        if self._chatstates.get(contact.jid) == State.INACTIVE:
            if was_paused:
                self.set_chatstate(contact, State.PAUSED)
            else:
                self.set_chatstate(contact, State.ACTIVE)

    @ensure_enabled
    def set_keyboard_activity(self, contact: types.ChatContactT) -> None:
        self._last_keyboard_activity[contact.jid] = time.time()

    def _remove_delay_timeout(self, contact: types.ChatContactT) -> None:
        timeout = self._delay_timeout_ids.get(contact.jid)
        if timeout is not None:
            GLib.source_remove(timeout)
            del self._delay_timeout_ids[contact.jid]

    def _remove_all_timeouts(self) -> None:
        for timeout in chain(
            self._delay_timeout_ids.values(), self._remote_composing_timeouts.values()
        ):
            GLib.source_remove(timeout)
        self._delay_timeout_ids.clear()
        self._remote_composing_timeouts.clear()

    def cleanup(self) -> None:
        BaseModule.cleanup(self)
        self._remove_all_timeouts()
        if self._timeout_id is not None:
            GLib.source_remove(self._timeout_id)
            self._timeout_id = None

        self._chatstates.clear()
        self._remote_chatstate.clear()
        self._last_keyboard_activity.clear()
        self._last_mouse_activity.clear()
        self._blocked = []



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging

from gi.repository import Adw
from gi.repository import Gtk
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.task import Task

from gajim.common import app
from gajim.common.i18n import _

from gajim.gtk.dropdown import GajimDropDown
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import ensure_not_destroyed
from gajim.gtk.util.misc import get_ui_string

log = logging.getLogger("gajim.gtk.preferences.archiving")


@Gtk.Template.from_string(string=get_ui_string("preference/archiving.ui"))
class ArchivingPreferences(Gtk.Box, SignalManager):
    __gtype_name__ = "ArchivingPreferences"

    _default_dropdown: GajimDropDown[str] = Gtk.Template.Child()
    _pref_view: Gtk.TreeView = Gtk.Template.Child()
    _jid_cell_renderer: Gtk.CellRendererText = Gtk.Template.Child()
    _pref_toggle_cell_renderer: Gtk.CellRendererToggle = Gtk.Template.Child()
    _spinner: Adw.Spinner = Gtk.Template.Child()
    _add: Gtk.Button = Gtk.Template.Child()
    _remove: Gtk.Button = Gtk.Template.Child()
    _save_button: Gtk.Button = Gtk.Template.Child()

    def __init__(self, account: str) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self.account = account
        self._client = app.get_client(account)
        self._destroyed = False

        self._default_dropdown.set_data(
            {
                "always": _("All Messages"),
                "roster": _("Messages from Contacts"),
                "never": _("No Messages"),
            }
        )

        self._preferences_store = Gtk.ListStore(str, bool)
        self._pref_view.set_model(self._preferences_store)

        self._connect(self._jid_cell_renderer, "edited", self._jid_edited)
        self._connect(self._pref_toggle_cell_renderer, "toggled", self._pref_toggled)
        self._connect(self._add, "clicked", self._on_add)
        self._connect(self._remove, "clicked", self._on_remove)
        self._connect(self._save_button, "clicked", self._on_save)

        self._set_mam_box_state(False)
        self._activate_spinner()

        self._client.get_module("MAM").request_preferences(
            callback=self._mam_prefs_received
        )

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Box.do_unroot(self)
        self._default_dropdown.run_destroy()
        self._destroyed = True
        app.check_finalize(self)

    @ensure_not_destroyed
    def _mam_prefs_received(self, task: Task) -> None:
        self._disable_spinner()
        self._set_mam_box_state(True)

        try:
            result = task.finish()
        except (StanzaError, MalformedStanzaError) as error:
            log.error("Failed to store archiving preferences: %s", error)
            return

        self._default_dropdown.select_key(result.default)

        self._preferences_store.clear()
        for jid in result.always:
            self._preferences_store.append([str(jid), True])

        for jid in result.never:
            self._preferences_store.append([str(jid), False])

    @ensure_not_destroyed
    def _mam_prefs_saved(self, task: Task) -> None:
        self._disable_spinner()
        self._set_mam_box_state(True)

        try:
            task.finish()
        except StanzaError as error:
            log.error("Failed to store archiving preferences: %s", error)
            return

    def _set_mam_box_state(self, state: bool) -> None:
        self.set_sensitive(state)

    def _jid_edited(
        self, _renderer: Gtk.CellRendererText, path: str, new_text: str
    ) -> None:
        iter_ = self._preferences_store.get_iter(path)
        self._preferences_store.set_value(iter_, 0, new_text)

    def _pref_toggled(self, _renderer: Gtk.CellRendererToggle, path: str):
        iter_ = self._preferences_store.get_iter(path)
        current_value = self._preferences_store[iter_][1]
        self._preferences_store.set_value(iter_, 1, not current_value)

    def _on_add(self, _button: Gtk.Button) -> None:
        self._preferences_store.append(["", False])

    def _on_remove(self, _button: Gtk.Button) -> None:
        rows = self._pref_view.get_selection().get_selected_rows()
        assert rows is not None
        mod, paths = rows
        for path in paths:
            iter_ = mod.get_iter(path)
            self._preferences_store.remove(iter_)

    def _on_save(self, _button: Gtk.Button) -> None:
        self._activate_spinner()
        self._set_mam_box_state(False)
        always: list[str] = []
        never: list[str] = []
        default = self._default_dropdown.get_selected_key()
        for item in self._preferences_store:
            jid, archive = item
            if archive:
                always.append(jid)
            else:
                never.append(jid)
        self._client.get_module("MAM").set_preferences(
            default, always, never, callback=self._mam_prefs_saved
        )

    def _activate_spinner(self) -> None:
        self._spinner.set_visible(True)

    def _disable_spinner(self) -> None:
        self._spinner.set_visible(False)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any

import logging

from gi.repository import Gio
from gi.repository import GLib

from gajim.common.util.version import gi_package_version

log = logging.getLogger("gajim.c.dbus")


class DBusService:
    def __init__(
        self,
        interface_info: Gio.DBusInterfaceInfo,
        object_path: str,
        bus: Gio.DBusConnection,
    ) -> None:
        self._interface_info = interface_info
        self._bus = bus
        self._object_path = object_path
        self._registration_id: int | None = None

    def register(self):
        if gi_package_version("GLib>=2.84.0"):
            register_method = self._bus.register_object_with_closures2
        else:
            register_method = self._bus.register_object

        try:
            self._registration_id = register_method(
                object_path=self._object_path,
                interface_info=self._interface_info,
                method_call_closure=self._on_method_call,
                get_property_closure=self._on_get_property,
            )
        except Exception as error:
            log.error(error)
            return

        assert self._registration_id

        self._interface_info.cache_build()

    def unregister(self):
        self._interface_info.cache_release()

        if self._registration_id is not None:
            try:
                self._bus.unregister_object(self._registration_id)
            except Exception as error:
                log.error(error)
            self._registration_id = None

    def _on_method_call(
        self,
        _connection: Gio.DBusConnection,
        _sender: str,
        _object_path: str,
        _interface_name: str,
        method_name: str,
        parameters: GLib.Variant,
        invocation: Gio.DBusMethodInvocation,
    ) -> None:
        method_info = self._interface_info.lookup_method(method_name)
        if method_info is None:
            log.warning("Unknown method name called: %s", method_name)
            return

        args = parameters.unpack()
        log.info("Method %s called on %s, args: %s", method_name, self, args)

        method = getattr(self, method_name)
        result = method(*args)
        out_arg_types = "".join([arg.signature for arg in method_info.out_args])
        return_value = None

        if method_info.out_args:
            return_value = GLib.Variant(f"({out_arg_types})", result)

        invocation.return_value(return_value)

    def _on_get_property(
        self,
        _connection: Gio.DBusConnection,
        _sender: str,
        _object_path: str,
        _interface_name: str,
        property_name: str,
    ) -> GLib.Variant | None:
        property_info = self._interface_info.lookup_property(property_name)
        if property_info is None:
            log.warning("Unknown property requested: %s", property_name)
            return

        value = getattr(self, property_name)
        log.info(
            "GetProperty %s called on %s, return value: %s", property_name, self, value
        )

        return GLib.Variant(property_info.signature, value)

    def emit_signal(self, signal_name: str, args: Any = None) -> None:
        signal_info = self._interface_info.lookup_signal(signal_name)
        if signal_info is None:
            raise ValueError(f"Unknown signal: {signal_name}")

        if len(signal_info.args) == 0:
            parameters = None
        else:
            arg_types = "".join([arg.signature for arg in signal_info.args])
            parameters = GLib.Variant(f"({arg_types})", args)

        log.info("Emit Signal %s on %s, args: %s", signal_name, self, args)

        try:
            self._bus.emit_signal(
                destination_bus_name=None,
                object_path=self._object_path,
                interface_name=self._interface_info.name,
                signal_name=signal_name,
                parameters=parameters,
            )
        except Exception as error:
            log.error(error)



# This file is part of Gajim.
#
# Gajim is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 3 only.
#
# Gajim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Gajim. If not, see <http://www.gnu.org/licenses/>.

# XEP-0425: Message Moderation

from __future__ import annotations

import datetime as dt

from nbxmpp import NodeProcessed
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common import types
from gajim.common.events import MessageModerated
from gajim.common.events import MessageReceived
from gajim.common.modules.base import BaseModule
from gajim.common.modules.message_util import get_chat_type_and_direction
from gajim.common.modules.message_util import UNKNOWN_MESSAGE
from gajim.common.storage.archive import models as mod
from gajim.common.storage.archive.const import MessageState
from gajim.common.storage.archive.const import MessageType
from gajim.common.structs import MUCData


class Moderations(BaseModule):
    def __init__(self, client: types.Client) -> None:
        BaseModule.__init__(self, client)

        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_moderation_message,
                typ="groupchat",
                ns=Namespace.FASTEN,
                priority=45,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_moderation_message,
                typ="groupchat",
                ns=Namespace.MESSAGE_RETRACT_1,
                priority=45,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_message_moderated_tombstone,
                typ="groupchat",
                ns=Namespace.MESSAGE_MODERATE,
                priority=45,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_message_moderated_tombstone,
                typ="groupchat",
                ns=Namespace.MESSAGE_RETRACT_1,
                priority=45,
            ),
        ]

    def _process_message_moderated_tombstone(
        self,
        _client: types.NBXMPPClient,
        stanza: Message,
        properties: MessageProperties,
    ) -> None:

        if not properties.is_moderation:
            return

        if not properties.is_mam_message:
            return

        assert properties.moderation is not None

        if not properties.moderation.is_tombstone:
            return

        assert properties.remote_jid is not None

        remote_jid = properties.remote_jid
        muc_data = self._client.get_module("MUC").get_muc_data(remote_jid)
        if muc_data is None:
            self._log.warning("Groupchat message from unknown MUC: %s", remote_jid)
            return

        is_occupant_id_supported = self._is_occupant_id_supported(properties)

        self._insert_tombstone(muc_data, properties, is_occupant_id_supported)

        raise NodeProcessed

    def _process_moderation_message(
        self,
        _client: types.NBXMPPClient,
        stanza: Message,
        properties: MessageProperties,
    ) -> None:

        if not properties.is_moderation:
            return

        is_occupant_id_supported = self._is_occupant_id_supported(properties)

        self._insert_moderation_message(properties, is_occupant_id_supported)

        raise NodeProcessed

    def _insert_moderation_message(
        self, properties: MessageProperties, is_occupant_id_supported: bool
    ) -> None:

        assert properties.moderation is not None

        moderator_nickname = self._get_moderator_nickname(properties)

        moderator_occupant_id = None
        if is_occupant_id_supported:
            moderator_occupant_id = properties.moderation.occupant_id

        remote_jid = properties.remote_jid
        assert remote_jid is not None

        occupant_data = None
        if moderator_occupant_id is not None:
            occupant_data = mod.Occupant(
                account_=self._account,
                remote_jid_=remote_jid.new_as_bare(),
                id=moderator_occupant_id,
                nickname=moderator_nickname,
                updated_at=properties.moderation.stamp,
            )

        moderation_data = mod.Moderation(
            account_=self._account,
            remote_jid_=remote_jid,
            occupant_=occupant_data,
            stanza_id=properties.moderation.stanza_id,
            by=properties.moderation.by,
            reason=properties.moderation.reason,
            timestamp=properties.moderation.stamp,
        )

        pk = app.storage.archive.insert_row(moderation_data, ignore_on_conflict=True)
        if pk == -1:
            return

        app.ged.raise_event(
            MessageModerated(
                account=self._account, jid=remote_jid, moderation=moderation_data
            )
        )

    def _insert_tombstone(
        self,
        muc_data: MUCData,
        properties: MessageProperties,
        is_occupant_id_supported: bool,
    ) -> None:

        assert properties.mam is not None

        remote_jid = properties.remote_jid
        assert remote_jid is not None
        assert properties.jid is not None

        message_occupant_id = None
        if is_occupant_id_supported:
            message_occupant_id = properties.occupant_id

        timestamp = dt.datetime.fromtimestamp(properties.mam.timestamp, dt.UTC)

        occupant_data = None
        if message_occupant_id is not None:
            occupant_data = mod.Occupant(
                account_=self._account,
                remote_jid_=remote_jid.new_as_bare(),
                id=message_occupant_id,
                nickname=properties.jid.resource,
                updated_at=timestamp,
            )

        m_type, direction = get_chat_type_and_direction(
            muc_data, self._client.get_own_jid(), properties
        )

        assert properties.id is not None

        message_data = mod.Message(
            account_=self._account,
            remote_jid_=remote_jid,
            resource=properties.jid.resource,
            type=m_type,
            direction=direction,
            timestamp=timestamp,
            state=MessageState.ACKNOWLEDGED,
            text=UNKNOWN_MESSAGE,
            id=properties.id,
            stanza_id=properties.mam.id,
            occupant_=occupant_data,
        )

        pk = app.storage.archive.insert_object(message_data)

        app.ged.raise_event(
            MessageReceived(
                account=self._account,
                jid=remote_jid,
                m_type=MessageType.GROUPCHAT,
                mam=properties.mam,
                pk=pk,
            )
        )

    def _is_occupant_id_supported(self, properties: MessageProperties) -> bool:
        assert properties.remote_jid is not None
        contact = self._client.get_module("Contacts").get_contact(
            properties.remote_jid, groupchat=True
        )
        return contact.supports(Namespace.OCCUPANT_ID)

    def _get_moderator_nickname(self, properties: MessageProperties) -> str | None:

        assert properties.moderation is not None
        if properties.moderation.by is None:
            return None
        return properties.moderation.by.resource



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only


from __future__ import annotations

from typing import Any
from typing import Generic
from typing import TypedDict
from typing import TypeVar

from datetime import datetime
from enum import IntEnum

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject

_K = TypeVar("_K")
_V = TypeVar("_V")


class Singleton(type):
    _instances: dict[Any, Any] = {}

    def __call__(cls, *args: Any, **kwargs: Any):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class SettingsAction(Gio.SimpleAction):
    def simple_bind_property(
        self, target: GObject.Object, target_property: str
    ) -> None:
        self.bind_property(
            "state",
            target,
            target_property,
            GObject.BindingFlags.SYNC_CREATE,
            transform_to=self._transform_to_ptype,
        )

    def change_state(self, value: GLib.Variant | Any) -> None:
        if not isinstance(value, GLib.Variant):
            v_type = self.get_state_type()
            assert v_type is not None
            value = GLib.Variant(v_type.dup_string(), value)
        super().change_state(value)

    @staticmethod
    def _transform_to_ptype(binding: GObject.Binding, variant: GLib.Variant) -> str:
        return variant.unpack()


class CacheResult(IntEnum):
    HIT = 0
    MISS = 1


class CacheItem(TypedDict, Generic[_V]):
    value: _V
    created: datetime


class TTLCache(Generic[_K, _V]):
    def __init__(self, *, ttl_seconds: int, extend_ttl_on_hit: bool):
        self._cache_items: dict[_K, CacheItem[_V]] = {}
        self._ttl_seconds = ttl_seconds
        self._extend_ttl_on_hit = extend_ttl_on_hit

    def get(self, key: _K) -> tuple[_V | None, CacheResult]:
        if self._is_expired(key):
            return None, CacheResult.MISS

        if self._extend_ttl_on_hit:
            self._cache_items[key]["created"] = datetime.now()
        return self._cache_items[key]["value"], CacheResult.HIT

    def add(self, key: _K, value: _V) -> None:
        self._cache_items[key] = CacheItem(value=value, created=datetime.now())

    def _is_expired(self, key: _K) -> bool:
        if key not in self._cache_items:
            return True

        time_diff = datetime.now() - self._cache_items[key]["created"]
        if time_diff.total_seconds() > self._ttl_seconds:
            del self._cache_items[key]
            return True

        return False

    def __contains__(self, key: _K) -> bool:
        return not self._is_expired(key)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import TypeVar

import base64
import dataclasses
import logging
from dataclasses import dataclass
from dataclasses import fields
from datetime import datetime

from gi.repository import GLib
from nbxmpp.const import Affiliation
from nbxmpp.const import PresenceShow
from nbxmpp.const import Role
from nbxmpp.modules.dataforms import SimpleDataForm
from nbxmpp.modules.security_labels import SecurityLabel
from nbxmpp.protocol import JID
from nbxmpp.protocol import Message
from nbxmpp.structs import EncryptionData
from nbxmpp.structs import HatData
from nbxmpp.structs import MucSubject
from nbxmpp.structs import PresenceProperties
from nbxmpp.util import generate_id

from gajim.common import types
from gajim.common.const import MUCJoinedState
from gajim.common.const import PresenceShowExt
from gajim.common.open_graph_parser import OpenGraphData
from gajim.common.storage.archive.const import MessageType
from gajim.common.util.datetime import convert_epoch_to_local_datetime
from gajim.common.util.datetime import utc_now

log = logging.getLogger("gajim.c.structs")

_T = TypeVar("_T")


class MUCData:
    def __init__(
        self,
        room_jid: str,
        nick: str,
        occupant_id: str | None,
        password: str | None,
        config: dict[str, Any] | None = None,
    ) -> None:

        self._room_jid = JID.from_string(room_jid)
        self._config = config
        self.nick = nick
        self.password = password
        self.occupant_id = occupant_id
        self.state = MUCJoinedState.NOT_JOINED
        # Message id of the captcha challenge
        self.captcha_id: str | None = None
        self.captcha_form: SimpleDataForm | None = None
        self.error: str | None = None
        self.error_text: str | None = None
        self.subject: MucSubject | None = None
        self.last_subject_timestamp: float | None = None

    @property
    def jid(self) -> JID:
        return self._room_jid

    @property
    def occupant_jid(self) -> JID:
        return self._room_jid.new_with(resource=self.nick)

    @property
    def config(self) -> dict[str, Any] | None:
        return self._config


@dataclass(slots=True)
class OutgoingMessage:
    account: str
    contact: types.ChatContactT
    text: dataclasses.InitVar[str | None] = None
    chatstate: str | None = None
    marker: tuple[str, str] | None = None
    mds_id: str | None = None
    sec_label: SecurityLabel | None = None
    control: Any | None = None
    correct_id: str | None = None
    reply_data: ReplyData | None = None
    open_graph_data: dict[str, OpenGraphData] | None = None
    reaction_data: tuple[str, set[str]] | None = None
    retraction_id: str | None = None
    oob_url: str | None = None
    play_sound: bool = True

    type: MessageType = dataclasses.field(init=False)
    message_id: str = dataclasses.field(init=False)
    timestamp: datetime = dataclasses.field(init=False)

    _text: str | None = dataclasses.field(init=False)
    _encryption_data: EncryptionData | None = dataclasses.field(
        init=False, default=None
    )
    _stanza: Any = dataclasses.field(init=False, default=None)

    def __post_init__(self, text: str | None) -> None:
        self._text = text
        self.timestamp = utc_now()
        self.message_id = generate_id()
        self.type = MessageType.from_str(self.contact.type_string)

    @property
    def is_groupchat(self) -> bool:
        return self.type == MessageType.GROUPCHAT

    @property
    def is_pm(self) -> bool:
        return self.type == MessageType.PM

    def get_text(self, with_fallback: bool = True) -> str | None:
        if not with_fallback:
            return self._text

        if self.reply_data is None:
            return self._text
        assert self._text is not None
        return f"{self.reply_data.fallback_text}{self._text}"

    def has_text(self) -> bool:
        return bool(self._text)

    def set_encryption(self, data: EncryptionData) -> None:
        self._encryption_data = data

    def get_encryption(self) -> EncryptionData | None:
        return self._encryption_data

    @property
    def is_encrypted(self) -> bool:
        return self._encryption_data is not None

    def set_stanza(self, stanza: Message) -> None:
        self._stanza = stanza

    def get_stanza(self) -> Message:
        return self._stanza


@dataclass(frozen=True)
class PresenceData:
    show: PresenceShow
    status: str
    priority: int
    idle_datetime: datetime | None
    available: bool
    hats: HatData | None

    @classmethod
    def from_presence(cls, properties: PresenceProperties) -> PresenceData:
        idle_datetime = None
        if properties.idle_timestamp is not None:
            idle_datetime = convert_epoch_to_local_datetime(properties.idle_timestamp)

        return cls(
            show=properties.show,
            status=properties.status,
            priority=properties.priority,
            idle_datetime=idle_datetime,
            available=properties.type.is_available,
            hats=properties.hats,
        )


UNKNOWN_PRESENCE = PresenceData(
    show=PresenceShowExt.OFFLINE,
    status="",
    priority=0,
    idle_datetime=None,
    available=False,
    hats=None,
)


@dataclass(frozen=True)
class MUCPresenceData:
    show: PresenceShow
    status: str
    idle_datetime: datetime | None
    available: bool
    affiliation: Affiliation
    role: Role
    real_jid: JID | None
    occupant_id: str | None
    hats: HatData | None

    @classmethod
    def from_presence(
        cls, properties: PresenceProperties, occupant_id: str | None
    ) -> MUCPresenceData:
        idle_datetime = None
        if properties.idle_timestamp is not None:
            idle_datetime = convert_epoch_to_local_datetime(properties.idle_timestamp)

        return cls(
            show=properties.show,
            status=properties.status,
            idle_datetime=idle_datetime,
            available=properties.type.is_available,
            affiliation=properties.muc_user.affiliation,
            role=properties.muc_user.role,
            real_jid=properties.muc_user.jid,
            occupant_id=occupant_id,
            hats=properties.hats,
        )


UNKNOWN_MUC_PRESENCE = MUCPresenceData(
    show=PresenceShowExt.OFFLINE,
    status="",
    idle_datetime=None,
    available=False,
    affiliation=Affiliation.NONE,
    role=Role.NONE,
    real_jid=None,
    occupant_id=None,
    hats=None,
)


ANNOTATION_TO_VARIANT = {
    "str": (str, "s"),
    "list[str]": (list, "as"),
    "Optional[str]": (str, "ms"),
    "bool": (bool, "b"),
    "int": (int, "i"),
    "JID": (JID, "s"),
    "Optional[JID]": (JID, "ms"),
    "Literal['chat', 'groupchat', 'pm']": (str, "s"),
}


class VariantMixin:
    _type_to_variant_funcs = {
        JID: str,
    }

    _variant_to_type_funcs = {
        "JID": JID.from_string,
    }

    def to_variant(self) -> GLib.Variant:
        __types = {}
        vdict = {}
        for field in fields(self):
            value = getattr(self, field.name)
            field_t, variant_str = ANNOTATION_TO_VARIANT[field.type]
            if value is None:
                vdict[field.name] = GLib.Variant(variant_str, value)
                continue

            if not isinstance(value, field_t):
                raise ValueError(
                    f"invalid type: {value} type {type(value)} is not a {field_t}"
                )

            conversion_func = self._type_to_variant_funcs.get(field_t)
            if conversion_func is not None:
                value = conversion_func(value)
                __types[field.name] = field_t.__name__

            vdict[field.name] = GLib.Variant(variant_str, value)

        if __types:
            vdict["__types"] = GLib.Variant("a{ss}", __types)
        return GLib.Variant("a{sv}", vdict)

    def serialize(self) -> str:
        variant = self.to_variant()
        variant_g_bytes = variant.get_data_as_bytes()
        variant_data = variant_g_bytes.get_data()
        assert variant_data is not None
        return base64.b64encode(variant_data).decode()

    @classmethod
    def from_serialized_string(
        cls: type[_T], serialized_data: str, variant_type: GLib.VariantType
    ) -> _T | None:
        try:
            data = base64.b64decode(serialized_data.encode())
            g_bytes = GLib.Bytes.new(data)
            variant = GLib.Variant.new_from_bytes(variant_type, g_bytes, True)
        except Exception as e:
            log.error("Could not decode serialized variant %s", e)
            return None

        return cls.from_variant(variant)

    @classmethod
    def from_variant(cls: type[_T], variant: GLib.Variant) -> _T:
        vdict = variant.unpack()
        __types = vdict.pop("__types", None)
        if __types is not None:
            for field_name, value_t_name in __types.items():
                value = vdict[field_name]
                conversion_func = cls._variant_to_type_funcs.get(value_t_name)
                if conversion_func is None:
                    raise ValueError(f"no conversion for: {value_t_name}")
                vdict[field_name] = conversion_func(value)
        return cls(**vdict)


@dataclass
class ReplyData:
    pk: int
    to: JID
    id: str
    thread_id: str | None
    fallback_start: int
    fallback_end: int
    fallback_text: str



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

import datetime as dt
from string import Template
from zoneinfo import ZoneInfo

from gi.repository import Gio
from nbxmpp.const import Affiliation
from nbxmpp.const import Chatstate
from nbxmpp.const import Role
from nbxmpp.modules.vcard4 import TzProperty
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common.const import GIO_TLS_ERRORS
from gajim.common.i18n import _
from gajim.common.i18n import ngettext
from gajim.common.i18n import p_


def format_idle_time(idle_time: dt.datetime) -> str:
    now = dt.datetime.now(dt.UTC)

    now_date = now.date()
    idle_date = idle_time.date()

    if idle_date == now_date:
        return idle_time.strftime(app.settings.get("time_format"))
    if idle_date == now_date - dt.timedelta(days=1):
        return _("Yesterday (%s)") % idle_time.strftime(app.settings.get("time_format"))
    if idle_date >= now_date - dt.timedelta(days=6):
        return idle_time.strftime(f"%a {app.settings.get('time_format')}")

    return idle_date.strftime(app.settings.get("date_format"))


def get_subscription_request_msg(account: str | None = None) -> str:
    if account is None:
        return _("I would like to add you to my contact list.")

    message = app.settings.get_account_setting(account, "subscription_request_msg")
    if message:
        return message

    contact = app.get_client(account).get_own_contact()

    message = _("Hello, I am $name. %s") % message
    return Template(message).safe_substitute({"name": contact.name})


def get_moderation_text(by: str | JID | None, reason: str | None) -> str:
    text = _("This message has been moderated")

    if by is not None:
        text = _("This message has been moderated by %s") % by

    if reason is None:
        return text

    reason_text = _("(reason: %s)") % reason
    return f"{text} {reason_text}"


def get_retraction_text(timestamp: dt.datetime) -> str:
    dt_format = app.settings.get("date_time_format")
    timestamp = timestamp.astimezone()
    timestamp_formatted = timestamp.strftime(dt_format)
    return _("This message has been retracted on {date}").format(
        date=timestamp_formatted
    )


def get_uf_sub(sub: str) -> str:
    if sub == "none":
        return p_("Contact subscription", "None")

    if sub == "to":
        return p_("Contact subscription", "To")

    if sub == "from":
        return p_("Contact subscription", "From")

    if sub == "both":
        return p_("Contact subscription", "Both")

    return p_("Contact subscription", "Unknown")


def get_uf_ask(ask: str | None) -> str:
    if ask is None:
        return p_("Contact subscription", "None")

    if ask == "subscribe":
        return p_("Contact subscription", "Subscribe")

    return ask


def get_uf_role(role: Role | str, plural: bool = False) -> str:
    """plural determines if you get Moderators or Moderator"""
    if not isinstance(role, str):
        role = role.value

    if role == "none":
        return p_("Group chat contact role", "None")
    if role == "moderator":
        if plural:
            return p_("Group chat contact role", "Moderators")
        return p_("Group chat contact role", "Moderator")
    if role == "participant":
        if plural:
            return p_("Group chat contact role", "Participants")
        return p_("Group chat contact role", "Participant")
    if role == "visitor":
        if plural:
            return p_("Group chat contact role", "Visitors")
        return p_("Group chat contact role", "Visitor")
    return ""


def get_uf_affiliation(affiliation: Affiliation | str, plural: bool = False) -> str:
    """Get a nice and translated affilition for muc"""
    if not isinstance(affiliation, str):
        affiliation = affiliation.value

    if affiliation == "none":
        return p_("Group chat contact affiliation", "None")
    if affiliation == "owner":
        if plural:
            return p_("Group chat contact affiliation", "Owners")
        return p_("Group chat contact affiliation", "Owner")
    if affiliation == "admin":
        if plural:
            return p_("Group chat contact affiliation", "Administrators")
        return p_("Group chat contact affiliation", "Administrator")
    if affiliation == "member":
        if plural:
            return p_("Group chat contact affiliation", "Members")
        return p_("Group chat contact affiliation", "Member")
    return ""


def get_uf_relative_time(date_time: dt.datetime, now: dt.datetime | None = None) -> str:
    # `date_time` must be an aware datetime object with timezone UTC
    # The parameter `now` is only used by unittests

    if date_time.tzinfo != dt.UTC:
        raise ValueError("param 'date_time' with timezone =! UTC")

    if now is None:  # used by unittest
        now = dt.datetime.now(dt.UTC).astimezone()
    else:
        if now.tzinfo != dt.UTC:
            raise ValueError("param 'now' with timezone =! UTC")

    date_time = date_time.astimezone()
    timespan = now - date_time

    if timespan < dt.timedelta(minutes=1):
        return _("Just now")
    if timespan < dt.timedelta(minutes=15):
        minutes = int(timespan.seconds / 60)
        return ngettext("%(number)s min ago", "%(number)s mins ago", minutes) % {
            "number": minutes
        }

    today = now.date()
    if date_time.date() == today:
        format_string = app.settings.get("time_format")
        return date_time.strftime(format_string)
    yesterday = now.date() - dt.timedelta(days=1)
    if date_time.date() == yesterday:
        return _("Yesterday")
    if timespan < dt.timedelta(days=7):  # this week
        return date_time.strftime("%a")  # weekday
    if timespan < dt.timedelta(days=365):  # this year
        return date_time.strftime("%b %d")
    return str(date_time.year)


def get_time_zone_string(prop: TzProperty) -> str:
    try:
        tzinfo = ZoneInfo(prop.value)
    except Exception:
        return ""

    t_format = app.settings.get("time_format")
    remote_dt_str = dt.datetime.now(tz=tzinfo).strftime(t_format)
    return f"{remote_dt_str} ({prop.value})"


def chatstate_to_string(chatstate: Chatstate | None) -> str:
    if chatstate is None:
        return ""

    if chatstate == Chatstate.ACTIVE:
        return ""

    if chatstate == Chatstate.COMPOSING:
        return _("is composing a message…")

    if chatstate in (Chatstate.INACTIVE, Chatstate.GONE):
        return _("is doing something else")

    if chatstate == Chatstate.PAUSED:
        return _("paused composing a message")

    raise ValueError(f"unknown value: {chatstate}")


def get_tls_error_phrases(tls_errors: set[Gio.TlsCertificateFlags]) -> list[str]:
    return [GIO_TLS_ERRORS[err] for err in tls_errors]
